<?php
if ( ! class_exists( 'Visualizer_Pro' ) ) {
	define( 'Visualizer_Pro_Name', 'visualizer' );
	define( 'Visualizer_Pro_Debug', false );

	/**
	 * Class Visualizer_Pro
	 *
	 * Bootstrap the pro features
	 */
	class Visualizer_Pro {

		/**
		 * Check if we have free version installed
		 *
		 * @var bool
		 */
		private $hasFreeInstalled = false;

		/**
		 * The default properties of the post.
		 *
		 * @var array
		 */
		private static $POST_DEFAULT_PROPERTIES = array(
			'post_title'    => 'string',
			'post_status'   => 'string',
			'comment_count' => 'int',
			'post_date'     => 'date',
			'post_modified' => 'date',
		);

		/**
		 * The list of aggregator functions.
		 *
		 * @var array
		 */
		private static $AGGREGATOR_FUNCTIONS = array(
			'sum',
			'avg',
			'count',
		);

		/**
		 *Action used for fetching the data
		 */
		const ACTION_FETCH_DATA = 'visualizer-fetch-data';

		/**
		 *Action used for fetching post type properties
		 */
		const ACTION_FETCH_POST_TYPE_PROPS = 'visualizer-fetch-post-type-props';

		/**
		 *Action used for fetching post type data
		 */
		const ACTION_FETCH_POST_TYPE_DATA = 'visualizer-fetch-post-type-data';

		/**
		 * Custom meta field used to store permissions.
		 */
		const CF_PERMISSIONS        = 'visualizer-permissions';

		/**
		 * Visualizer_Pro constructor.
		 */
		function __construct() {
			register_activation_hook( VISUALIZER_PRO_BASEFILE, array( $this, 'activate' ) );
			register_deactivation_hook( VISUALIZER_PRO_BASEFILE, array( $this, 'deactivate' ) );
			$this->loadHooks();
		}

		/**
		 * Load all pro hooks
		 */
		private function loadHooks() {
			add_action( 'admin_init', array( $this, 'adminInit' ) );
			add_action( 'visualizer_pro_debug', array( __CLASS__, 'writeDebug' ) );
			add_action( 'wp_ajax_' . self::ACTION_FETCH_DATA, array( $this, 'fetchData' ) );
			add_action( 'wp_ajax_' . self::ACTION_FETCH_POST_TYPE_PROPS, array( $this, 'fetch_post_type_properties' ) );
			add_action( 'wp_ajax_' . self::ACTION_FETCH_POST_TYPE_DATA, array( $this, 'fetch_post_type_data' ) );
			add_action( 'visualizer_schedule_import', array( $this, 'updateScheduledChart' ) );
			add_action( 'visualizer_pro_handle_tab', array( $this, 'handlePermissionsTab' ), 10, 2 );
			add_action( 'visualizer_pro_frontend_load_resources', array( $this, 'loadFrontendResources' ) );
			add_action( 'visualizer_pro_new_chart_defaults', array( $this, 'createNewChart' ), 10, 1 );

			add_filter( 'visualizer_pro_chart_types', array( $this, 'chartTypes' ) );
			add_filter( 'visualizer_pro_chart_type_sidebar', array( $this, 'chartTypeSidebar' ), 10, 2 );
			add_filter( 'visualizer_pro_upsell', array( $this, 'upsell' ), 9, 2 );
			add_filter( 'visualizer_pro_upsell_class', array( $this, 'upsell' ), 9, 2 );
			add_filter( 'visualizer_pro_handle_chart_data', array( $this, '_handleChartData' ), 10, 2 );
			add_filter( 'visualizer_pro_chart_schedule', array( $this, 'updateSchedule' ), 10, 3 );
			add_filter( 'visualizer_pro_remove_schedule', array( $this, 'deleteSchedule' ), 10, 1 );
			add_filter( 'visualizer_is_business', array( $this, 'visualizer_is_business' ), 10, 1 );
			add_filter( 'visualizer_pro_logger_data', array( $this, 'getLoggerData' ), 10, 1 );
			add_filter( 'visualizer_pro_inline_css', array( $this, 'getInlineCSS' ) );
			add_filter( 'visualizer_pro_add_actions', array( $this, 'addActions' ), 10, 2 );
			add_filter( 'visualizer_action_buttons', array( $this, 'addActionButtons' ), 10, 1 );
			add_filter( 'visualizer_action_data', array( $this, 'handleAction' ), 10, 5 );
			add_filter( 'visualizer_pro_show_chart', array( $this, 'checkViewChartPermission' ), 10, 2 );
			add_filter( 'visualizer_pro_get_permissions', array( $this, 'getPermissions' ), 10, 2 );
			add_filter( 'visualizer_pro_get_permissions_data', array( $this, 'getPermissionsDataFilter' ), 10, 2 );

			add_action( 'after_setup_theme', array( $this, 'loadDependencies' ), 999 );
		}

		/**
		 * Get the permissions meta data for the chart.
		 *
		 * @access public
		 * @param null $default Null.
		 * @param int  $chart_id The id of the chart.
		 */
		public function getPermissions( $default = null, $chart_id ) {
			return get_post_meta( $chart_id, self::CF_PERMISSIONS, true );
		}

		/**
		 * Called while creating a new chart.
		 *
		 * @access public
		 * @param int $chart_id The id of the chart.
		 */
		public function createNewChart( $chart_id ) {
			update_post_meta(
				$chart_id, self::CF_PERMISSIONS, array(
					'permissions'   => array(
						'read'          => 'all',
						'edit'          => 'roles',
						'edit-specific' => array( 'administrator' ),
					),
				)
			);
		}

		/**
		 * Loads resources for the front end.
		 *
		 * @access public
		 */
		public function loadFrontendResources() {
			if ( ! apply_filters( 'visualizer_is_business', false ) ) {
				return false;
			}

			wp_register_script( 'visualizer-handsontable', Visualizer_Pro_ABSURL . 'vendor/handsontable/dist/handsontable.full.min.js', array( 'jquery' ), VISUALIZER_PRO_VERSION, true );
			wp_register_script( 'visualizer-magnific', '//cdnjs.cloudflare.com/ajax/libs/magnific-popup.js/1.1.0/jquery.magnific-popup.min.js', array( 'visualizer-handsontable' ), VISUALIZER_PRO_VERSION, true );

			wp_register_script(
				'visualizer-pro-render-edit', Visualizer_Pro_ABSURL . 'inc/js/render.js', array(
					'visualizer-magnific',
				), VISUALIZER_PRO_VERSION, true
			);
			wp_localize_script(
				'visualizer-pro-render-edit', 'visualizer2', array(
					'locale' => get_locale(),
					'types' => $this->getAllowedTypes(),
					'i10n'  => array(
						'save'      => esc_html__( 'Save', 'visualizer' ),
						'cancel'    => esc_html__( 'Cancel', 'visualizer' ),
					),
					'rest_url'      => rest_url( 'visualizer/v' . VISUALIZER_REST_VERSION . '/action/#id#/#type#/' ),
				)
			);

			wp_register_script(
				'visualizer-pro-render', Visualizer_Pro_ABSURL . 'inc/js/render.js', array(), VISUALIZER_PRO_VERSION, true
			);
			wp_localize_script(
				'visualizer-pro-render', 'visualizer2', array(
					'types' => $this->getAllowedTypes(),
					'i10n'  => array(
						'save'      => esc_html__( 'Save', 'visualizer' ),
						'cancel'    => esc_html__( 'Cancel', 'visualizer' ),
					),
					'rest_url'      => rest_url( 'visualizer/v' . VISUALIZER_REST_VERSION . '/action/#id#/#type#/' ),
				)
			);

			wp_register_style( 'visualizer-handsontable', Visualizer_Pro_ABSURL . 'vendor/handsontable/dist/handsontable.full.min.css', array(), VISUALIZER_PRO_VERSION );
			wp_register_style( 'visualizer-magnific', '//cdnjs.cloudflare.com/ajax/libs/magnific-popup.js/1.1.0/magnific-popup.min.css', array( 'visualizer-handsontable' ), VISUALIZER_PRO_VERSION );
			wp_register_style( 'visualizer-pro-front-edit', Visualizer_Pro_ABSURL . 'inc/css/front.css', array( 'visualizer-magnific' ), VISUALIZER_PRO_VERSION );
			wp_register_style( 'visualizer-pro-front', Visualizer_Pro_ABSURL . 'inc/css/front.css', array(), VISUALIZER_PRO_VERSION );
		}

		/**
		 * Check whether to show the chart or not.
		 *
		 * @access public
		 * @param bool $show The default action (true).
		 * @param int  $chart_id The id of the chart.
		 */
		public function checkViewChartPermission( $show, $chart_id ) {
			return $this->checkChartPermissionFor( 'read', $show, $chart_id );
		}

		/**
		 * Check whether the chart has a certain permission or not.
		 *
		 * @access private
		 * @param string $type The type of permission.
		 * @param bool   $show The default action (true).
		 * @param int    $chart_id The id of the chart.
		 */
		private function checkChartPermissionFor( $type, $show, $chart_id ) {
			if ( ! apply_filters( 'visualizer_is_business', false ) ) {
				return $show;
			}

			$permissions    = get_post_meta( $chart_id, self::CF_PERMISSIONS, true );
			if ( $permissions ) {
				if ( isset( $permissions['permissions'] ) && isset( $permissions['permissions'][ $type ] ) ) {
					$show       = false;
					switch ( $permissions['permissions'][ $type ] ) {
						case 'all':
							$show       = true;
							break;
						case 'users':
							$allowed    = isset( $permissions['permissions'][ $type . '-specific' ] ) ? $permissions['permissions'][ $type . '-specific' ] : array();
							$show       = is_user_logged_in() && in_array( get_current_user_id(), $allowed );
							break;
						case 'roles':
							global $current_user;
							$roles      = $current_user->roles;
							$role       = array_shift( $roles );
							$allowed    = isset( $permissions['permissions'][ $type . '-specific' ] ) ? $permissions['permissions'][ $type . '-specific' ] : array();
							$show       = is_user_logged_in() && in_array( $role, $allowed );
							break;
					}
				}
			}
			return $show;
		}


		/**
		 * Handles the custom actions.
		 *
		 * @access public
		 * @param array               $data The default data (null).
		 * @param int                 $chart_id The id of the chart.
		 * @param string              $type The type of action.
		 * @param WP_REST_Request     $params The parameters of the request.
		 * @param Visualizer_Frontend $frontend_module The front end module instance.
		 */
		public function handleAction( $data, $chart_id, $type, WP_REST_Request $params, $frontend_module ) {
			if ( ! apply_filters( 'visualizer_is_business', false ) ) {
				return $data;
			}

			switch ( $type ) {
				case 'edit':
					$chart      = get_post( $chart_id );
					$chart_type = get_post_meta( $chart->ID, Visualizer_Plugin::CF_CHART_TYPE, true );
					$series     = apply_filters( Visualizer_Plugin::FILTER_GET_CHART_SERIES, get_post_meta( $chart->ID, Visualizer_Plugin::CF_SERIES, true ), $chart->ID, $chart_type );
					$content    = apply_filters( Visualizer_Plugin::FILTER_GET_CHART_DATA, unserialize( $chart->post_content ), $chart->ID, $chart_type );
					$data       = array(
						'series'    => $series,
						'data'      => $content,
					);
					break;
				case 'save':
					$source     = $this->_handleChartData( $params['data'] );
					if ( $source ) {
						if ( $source->fetch() ) {
							$chart  = get_post( $chart_id );
							$chart->post_content = $source->getData();
							wp_update_post( $chart->to_array() );
							update_post_meta( $chart_id, Visualizer_Plugin::CF_SERIES, $source->getSeries() );
							update_post_meta( $chart_id, Visualizer_Plugin::CF_SOURCE, $source->getSourceName() );
							update_post_meta( $chart_id, Visualizer_Plugin::CF_DEFAULT_DATA, 0 );
							apply_filters( Visualizer_Plugin::FILTER_UNDO_REVISIONS, $chart_id, false );
						}
					}
					break;
			}

			return $data;
		}

		/**
		 * Adds the custom actions.
		 *
		 * @access public
		 * @param array $actions The default actions.
		 */
		public function addActionButtons( $actions ) {
			if ( ! apply_filters( 'visualizer_is_business', false ) ) {
				return $actions;
			}

			$actions['edit']    = esc_html__( 'Edit', 'visualizer' );
			return $actions;
		}

		/**
		 * Adds the custom actions names.
		 *
		 * @access public
		 * @param array $actions The default actions, as enabled in the settings.
		 * @param int   $chart_id The id of the chart.
		 */
		public function addActions( $actions, $chart_id ) {
			if ( $this->checkChartPermissionFor( 'edit', false, $chart_id ) ) {
				$actions['edit']    = 'edit';
				wp_enqueue_style( 'visualizer-pro-front-edit' );
				wp_enqueue_script( 'visualizer-pro-render-edit' );
			} else {
				wp_enqueue_style( 'visualizer-pro-front' );
				wp_enqueue_script( 'visualizer-pro-render' );
			}
			return $actions;
		}

		/**
		 * Handles the permissions data.
		 *
		 * @access public
		 * @param string  $tab The name of the tab.
		 * @param WP_Post $chart The chart object.
		 */
		public function handlePermissionsTab( $tab, $chart ) {
			if ( apply_filters( 'visualizer_is_business', false ) && 'permissions' === $tab ) {
				update_post_meta( $chart->ID, self::CF_PERMISSIONS, $_POST );
			}

			$render       = new Visualizer_Render_Page_Send();
			$render->text = sprintf( '[visualizer id="%d"]', $chart->ID );
			wp_iframe( array( $render, 'render' ) );
			return;
		}

		/**
		 * Returns the permissions data.
		 *
		 * @access public
		 * @params string $default Null.
		 * @params string $type The type of data to fetch for.
		 */
		public function getPermissionsDataFilter( $default, $type ) {
			return $this->getPermissionsData( $type );
		}

		/**
		 * Gets the data for users/roles.
		 *
		 * @params string $type The type of data to fetch for.
		 * @access private
		 */
		private function getPermissionsData( $type ) {
			$options    = array();
			switch ( $type ) {
				case 'users':
					$query  = new WP_User_Query(
						array(
							'number'        => 1000,
							'orderby'       => 'display_name',
							'fields'        => array( 'ID', 'display_name' ),
							'count_total'   => false,
						)
					);
					$users  = $query->get_results();
					if ( ! empty( $users ) ) {
						foreach ( $users as $user ) {
							$options[ $user->ID ] = $user->display_name;
						}
					}
					break;
				case 'roles':
					$roles  = get_editable_roles();
					if ( ! empty( $roles ) ) {
						foreach ( $roles as $name => $info ) {
							$options[ $name ] = $name;
						}
					}
					break;
			}
			return $options;
		}

		/**
		 * Gets the data for users/roles.
		 *
		 * @access friendly
		 */
		function fetch_permissions_data() {
			if ( ! defined( 'Visualizer_Plugin::ACTION_FETCH_PERMISSIONS_DATA' ) ) {
				return;
			}
			global $wpdb;
			$nonce  = isset( $_POST['nonce'] ) ? wp_verify_nonce( $_POST['nonce'], Visualizer_Plugin::ACTION_FETCH_PERMISSIONS_DATA ) : '';
			$type   = isset( $_POST['type'] ) ? $_POST['type'] : '';
			if ( $nonce && $type ) {
				$this->_sendResponse(
					array(
						'success' => true,
						'data'    => $this->getPermissionsData( $type ),
					)
				);
				wp_die();
			}
		}

		/**
		 * Generates the inline CSS to apply to the chart and adds these classes to the settings.
		 *
		 * @access public
		 * @param array $array The array (original inline css, id of the chart, settings of the chart) passed by reference.
		 */
		public function getInlineCSS( &$array ) {
			$css        = $array[0];
			$id         = $array[1];
			$settings   = $array[2];

			if ( ! isset( $settings['customcss'] ) ) {
				return;
			}

			$classes    = array();
			$css        = '<style type="text/css" name="visualizer-custom-css" id="css-' . $id . '">';
			foreach ( $settings['customcss'] as $name => $element ) {
				$attributes = array();
				foreach ( $element as $property => $value ) {
					$attributes[]   = $this->handleCSSProperty( $property, $value );
				}
				$class_name = $id . $name;
				$properties = implode( '; ', array_filter( $attributes ) );
				if ( ! empty( $properties ) ) {
					$css    .= '.' . $class_name . ' {' . $properties . ';}';
					$classes[ $name ] = $class_name;
				}
			}
			$css        .= '</style>';

			$settings['cssClassNames']  = $classes;

			$array[0]   = $css;
			$array[2]   = $settings;
		}

		/**
		 * Handles CSS properties that might need special syntax.
		 *
		 * @access private
		 * @param string $property The name of the css property.
		 * @param string $value The value of the css property.
		 */
		private function handleCSSProperty( $property, $value ) {
			if ( empty( $property ) || empty( $value ) ) {
				return '';
			}

			switch ( $property ) {
				case 'transform':
					$value  = 'rotate(' . $value . 'deg)';
					break;
			}
			return $property . ': ' . $value;
		}

		/**
		 * Fetches the SDK logger data.
		 *
		 * @param array $data The default data that needs to be sent.
		 *
		 * @access public
		 */
		public function getLoggerData( $data ) {
			return apply_filters(
				'visualizer_get_chart_counts', array(
					'wordpress_filters' => 'visualizer-filter-config',
				)
			);
		}

		/**
		 * Add/update the chart schedule
		 *
		 * @param int    $chart_id chart id.
		 * @param string $url the url.
		 * @param int    $hours how many hours.
		 */
		function updateSchedule( $chart_id, $url, $hours ) {
			update_post_meta( $chart_id, Visualizer_Plugin::CF_CHART_URL, $url );
			update_post_meta( $chart_id, Visualizer_Plugin::CF_CHART_SCHEDULE, $hours );
			$schedules              = get_option( Visualizer_Plugin::CF_CHART_SCHEDULE, array() );
			$schedules[ $chart_id ] = time() + $hours * HOUR_IN_SECONDS;
			update_option( Visualizer_Plugin::CF_CHART_SCHEDULE, $schedules );

			return null;
		}

		/**
		 * Delete the chart from the schedule
		 *
		 * @param int $chart_id chart id.
		 */
		function deleteSchedule( $chart_id ) {
			delete_post_meta( $chart_id, Visualizer_Plugin::CF_CHART_SCHEDULE );
			delete_post_meta( $chart_id, Visualizer_Plugin::CF_CHART_URL );
			$schedules = get_option( Visualizer_Plugin::CF_CHART_SCHEDULE, array() );
			if ( $schedules ) {
				unset( $schedules[ $chart_id ] );
			}
			update_option( Visualizer_Plugin::CF_CHART_SCHEDULE, $schedules );
		}

		/**
		 * Activate the plugin
		 */
		function activate() {
			wp_clear_scheduled_hook( 'visualizer_schedule_import' );
			wp_schedule_event( strtotime( 'midnight' ) - get_option( 'gmt_offset' ) * HOUR_IN_SECONDS, apply_filters( 'visualizer_chart_schedule_interval', 'hourly' ), 'visualizer_schedule_import' );
		}

		/**
		 * Deactivate the plugin
		 */
		function deactivate() {
			wp_clear_scheduled_hook( 'visualizer_schedule_import' );
		}

		/**
		 * Update the charts that are scheduled for now
		 */
		function updateScheduledChart() {
			$schedules = get_option( Visualizer_Plugin::CF_CHART_SCHEDULE, array() );
			if ( ! $schedules ) {
				return;
			}
			if ( ! defined( 'VISUALIZER_DO_NOT_DIE' ) ) {
				// define this so that the ajax call does not die
				// this means that if the new version of pro and the old version of free are installed, only the first chart will be updated
				define( 'VISUALIZER_DO_NOT_DIE', true );
			}
			$new_schedules = array();
			$now           = time();
			foreach ( $schedules as $chart_id => $time ) {
				$new_schedules[ $chart_id ] = $time;
				if ( $time > $now ) {
					continue;
				}
				// run it
				$_GET['nonce']           = wp_create_nonce();
				$_GET['chart']           = $chart_id;
				$_POST['remote_data']    = get_post_meta( $chart_id, Visualizer_Plugin::CF_CHART_URL, true );
				$_POST['vz-import-time'] = get_post_meta( $chart_id, Visualizer_Plugin::CF_CHART_SCHEDULE, true );
				do_action( 'wp_ajax_' . Visualizer_Plugin::ACTION_UPLOAD_DATA );
				apply_filters( Visualizer_Plugin::FILTER_UNDO_REVISIONS, $chart_id, false );
				$hours                      = get_post_meta( $chart_id, Visualizer_Plugin::CF_CHART_SCHEDULE, true );
				$new_schedules[ $chart_id ] = time() + $hours * HOUR_IN_SECONDS;
			}
			update_option( Visualizer_Plugin::CF_CHART_SCHEDULE, $new_schedules );
		}

		/**
		 * Return the upsell html/class
		 *
		 * @param string $dummy dummy.
		 * @param string $feature feature.
		 *
		 * @return string
		 */
		function upsell( $dummy, $feature = null ) {
			if ( ! $feature || ( ( $feature == 'schedule-chart' || $feature == 'chart-permissions' ) && apply_filters( 'visualizer_is_business', false ) ) ) {
				return '';
			}

			return $dummy;
		}

		/**
		 * Build the sidebar with pro features
		 *
		 * @param mixed $sidebar Sidebar to load.
		 * @param mixed $data Chart specific info.
		 *
		 * @return mixed
		 */
		function chartTypeSidebar( $sidebar, $data ) {
			if ( file_exists( Visualizer_Pro_PATH . '/inc/types/' . ucwords( $data['type'] ) . '.php' ) ) {
				require_once Visualizer_Pro_PATH . '/inc/types/' . ucwords( $data['type'] ) . '.php';
				$type    = ucwords( $data['type'] );
				$sidebar = new $type( $data['settings'] );
			}

			return $sidebar;
		}

		/**
		 * Add chart types from the pro version
		 *
		 * @param array $types Array of available charts.
		 *
		 * @return mixed
		 */
		function chartTypes( $types ) {
			// convert old charts to new values;
			if ( isset( $types['line']['name'] ) ) {
				$types['table']    = array(
					'name'    => esc_html__( 'Table', 'visualizer' ),
					'enabled' => true,
				);
				$types['timeline'] = array(
					'name'    => esc_html__( 'Timeline', 'visualizer' ),
					'enabled' => true,
				);
				$types['combo']    = array(
					'name'    => esc_html__( 'Combo', 'visualizer' ),
					'enabled' => true,
				);
			} else {
				$types['table']    = esc_html__( 'Table', 'visualizer' );
				$types['timeline'] = esc_html__( 'Timeline', 'visualizer' );
				$types['combo']    = esc_html__( 'Combo', 'visualizer' );
			}

			return $types;
		}

		/**
		 * Loads the plugin deps
		 */
		function loadDependencies() {
			if ( ! function_exists( 'tgmpa' ) ) {
				include_once Visualizer_Pro_PATH . '/lib/tgmpa/tgm-plugin-activation/class-tgm-plugin-activation.php';
			}

			if ( function_exists( 'tgmpa' ) ) {
				add_action( 'tgmpa_register', array( $this, 'tgmpa_register' ) );
			}
		}

		/**
		 * Initialize TGM.
		 */
		public function tgmpa_register() {
			$plugins = array(
				array(
					'name'     => 'Visualizer Lite',
					'slug'     => 'visualizer',
					'required' => true,
				),
			);
			$config  = array(
				'id'           => 'visualizer',
				// Unique ID for hashing notices for multiple instances of TGMPA.
				'default_path' => '',
				// Default absolute path to bundled plugins.
				'menu'         => 'tgmpa-install-plugins',
				// Menu slug.
				'parent_slug'  => 'plugins.php',
				// Parent menu slug.
				'capability'   => 'manage_options',
				// Capability needed to view plugin install page, should be a capability associated with the parent menu used.
				'has_notices'  => true,
				// Show admin notices or not.
				'dismissable'  => true,
				// If false, a user cannot dismiss the nag message.
				'dismiss_msg'  => '',
				// If 'dismissable' is false, this message will be output at top of nag.
				'is_automatic' => false,
				// Automatically activate plugins after installation or not.
				'message'      => '',
				// Message to output right before the plugins table.
				'strings'      => array(
					'page_title'                      => __( 'Install Required Plugins', 'visualizer' ),
					'menu_title'                      => __( 'Install Plugins', 'visualizer' ),
					/* translators: %s: plugin name. */
					'installing'                      => __( 'Installing Plugin: %s', 'visualizer' ),
					/* translators: %s: plugin name. */
					'updating'                        => __( 'Updating Plugin: %s', 'visualizer' ),
					'oops'                            => __( 'Something went wrong with the plugin API.', 'visualizer' ),
					/* translators: 1: plugin name(s). */
					'notice_can_install_required'     => _n_noop(
						'Visualizer requires the following plugin: %1$s.',
						'Visualizer requires the following plugins: %1$s.',
						'visualizer'
					),
					/* translators: 1: plugin name(s). */
					'notice_can_install_recommended'  => _n_noop(
						'Visualizer recommends the following plugin: %1$s.',
						'Visualizer recommends the following plugins: %1$s.',
						'visualizer'
					),
					/* translators: 1: plugin name(s). */
					'notice_ask_to_update'            => _n_noop(
						'The following plugin needs to be updated to its latest version to ensure maximum compatibility with Visualizer: %1$s.',
						'The following plugins need to be updated to their latest version to ensure maximum compatibility with Visualizer: %1$s.',
						'visualizer'
					),
					/* translators: 1: plugin name(s). */
					'notice_ask_to_update_maybe'      => _n_noop(
						'There is an update available for: %1$s.',
						'There are updates available for the following plugins: %1$s.',
						'visualizer'
					),
					/* translators: 1: plugin name(s). */
					'notice_can_activate_required'    => _n_noop(
						'The following required plugin is currently inactive: %1$s.',
						'The following required plugins are currently inactive: %1$s.',
						'visualizer'
					),
					/* translators: 1: plugin name(s). */
					'notice_can_activate_recommended' => _n_noop(
						'The following recommended plugin is currently inactive: %1$s.',
						'The following recommended plugins are currently inactive: %1$s.',
						'visualizer'
					),
					'install_link'                    => _n_noop(
						'Begin installing plugin',
						'Begin installing plugins',
						'visualizer'
					),
					'update_link'                     => _n_noop(
						'Begin updating plugin',
						'Begin updating plugins',
						'visualizer'
					),
					'activate_link'                   => _n_noop(
						'Begin activating plugin',
						'Begin activating plugins',
						'visualizer'
					),
					'return'                          => __( 'Return to Required Plugins Installer', 'visualizer' ),
					'plugin_activated'                => __( 'Plugin activated successfully.', 'visualizer' ),
					'activated_successfully'          => __( 'The following plugin was activated successfully:', 'visualizer' ),
					/* translators: 1: plugin name. */
					'plugin_already_active'           => __( 'No action taken. Plugin %1$s was already active.', 'visualizer' ),
					/* translators: 1: plugin name. */
					'plugin_needs_higher_version'     => __( 'Plugin not activated. A higher version of %s is needed for Visualizer. Please update the plugin.', 'visualizer' ),
					/* translators: 1: dashboard link. */
					'complete'                        => __( 'All plugins installed and activated successfully. %1$s', 'visualizer' ),
					'dismiss'                         => __( 'Dismiss this notice', 'visualizer' ),
					'notice_cannot_install_activate'  => __( 'There are one or more required or recommended plugins to install, update or activate.', 'visualizer' ),
					'contact_admin'                   => __( 'Please contact the administrator of this site for help.', 'visualizer' ),
					'nag_type'                        => '',
					// Determines admin notice type - can only be one of the typical WP notice classes, such as 'updated', 'update-nag', 'notice-warning', 'notice-info' or 'error'. Some of which may not work as expected in older WP versions.
				),
			);
			tgmpa( $plugins, $config );
		}

		/**
		 * Init the admin
		 */
		function adminInit() {
			$this->hasFreeInstalled = defined( 'VISUALIZER_BASEFILE' );
			if ( ! defined( 'Visualizer_Plugin::ACTION_FETCH_PERMISSIONS_DATA' ) ) {
				return;
			}
			add_action( 'wp_ajax_' . Visualizer_Plugin::ACTION_FETCH_PERMISSIONS_DATA, array( $this, 'fetch_permissions_data' ) );
		}

		/**
		 * Ads scripts and styles
		 */
		public function _addScriptsAndStyles() {
			if ( ! $this->hasFreeInstalled ) {
				return;
			}
			wp_register_script( 'visualizer-handsontable', Visualizer_Pro_ABSURL . 'vendor/handsontable/dist/handsontable.full.min.js', array(), VISUALIZER_PRO_VERSION, true );
			wp_register_script( 'visualizer-chosen', Visualizer_Pro_ABSURL . 'inc/js/chosen.jquery.min.js', array(), VISUALIZER_PRO_VERSION );
			wp_register_script(
				'visualizer-editor', Visualizer_Pro_ABSURL . 'inc/js/editor.js', array(
					'visualizer-handsontable',
					'visualizer-chosen',
				), VISUALIZER_PRO_VERSION, true
			);
			wp_register_style( 'visualizer-handsontable', Visualizer_Pro_ABSURL . 'vendor/handsontable/dist/handsontable.full.min.css', array(), VISUALIZER_PRO_VERSION );
			wp_register_style( 'visualizer-chosen', Visualizer_Pro_ABSURL . 'inc/css/chosen.min.css', array(), VISUALIZER_PRO_VERSION );
			wp_register_style( 'visualizer-main', Visualizer_Pro_ABSURL . 'inc/css/main.css', array(), VISUALIZER_PRO_VERSION );
		}

		/**
		 * Enqueue the admin script and styles
		 *
		 * @param mixed $data Data to output for js scripts.
		 */
		public function _enqueueScriptsAndStyles( $data ) {
			if ( ! $this->hasFreeInstalled ) {
				return;
			}
			// support for old free
			$types = $this->getAllowedTypes();
			wp_enqueue_script( 'visualizer-handsontable' );
			wp_enqueue_style( 'visualizer-handsontable' );
			wp_enqueue_script( 'visualizer-editor' );
			wp_enqueue_script( 'visualizer-chosen' );
			wp_enqueue_style( 'visualizer-chosen' );
			wp_enqueue_style( 'visualizer-main' );
			wp_localize_script(
				'visualizer-editor', 'visualizer1', array(
					'locale'            => get_locale(),
					'data'                 => $data,
					'types'                => $types,
					'max_selected_options' => 2,
					'l10n'                 => array(
						'invalid_source'      => esc_html__( 'You have entered invalid URL. Please, insert proper URL.', 'visualizer' ),
						'loading'             => esc_html__( 'Loading...', 'visualizer' ),
						'filter_config_error' => esc_html__( 'Please check the filters you have configured.', 'visualizer' ),
					),
					'ajax'                 => array(
						'url'     => admin_url( 'admin-ajax.php' ),
						'nonces'  => array(
							'filter_get_props' => wp_create_nonce( self::ACTION_FETCH_POST_TYPE_PROPS ),
							'filter_get_data'  => wp_create_nonce( self::ACTION_FETCH_POST_TYPE_DATA ),
						),
						'actions' => array(
							'filter_get_props' => self::ACTION_FETCH_POST_TYPE_PROPS,
							'filter_get_data'  => self::ACTION_FETCH_POST_TYPE_DATA,
						),
					),
					'filter'    => defined( 'Visualizer_Plugin::CF_SOURCE_FILTER' ) ? Visualizer_Plugin::CF_SOURCE_FILTER : '',
				)
			);
		}

		/**
		 * Handle chart import data
		 *
		 * @param mixed $data Chart data.
		 * @param mixed $dumb empty always.
		 *
		 * @return Visualizer_Source_Csv
		 */
		function _handleChartData( $data, $dumb = null ) {
			$tmpfile = tempnam( get_temp_dir(), Visualizer_Plugin::NAME );
			$handle  = fopen( $tmpfile, 'w' );
			$arr     = json_decode( stripslashes( trim( $data ) ), true );
			foreach ( $arr as $array ) {
				$data = array_filter( $array, array( __CLASS__, 'chart_array_filter' ) );
				if ( $data ) {
					fputcsv( $handle, $data );
				}
			}
			$source = new Visualizer_Source_Csv( $tmpfile );
			fclose( $handle );

			// unlink($tmpfile);
			return $source;
		}

		/**
		 * Filter the chart array
		 *
		 * @param string $val Cell value.
		 *
		 * @return bool
		 */
		private static function chart_array_filter( $val ) {
			return strlen( trim( $val ) ) != 0;
		}

		/**
		 * Adds the editor
		 */
		public function _addEditor( $chart_id = null ) {
			if ( ! $this->hasFreeInstalled ) {
				return;
			}
			?>
			<div id="chart-editor" class="visualizer-editor-lhs" style="display: none"></div>
			<?php
		}

		/**
		 * Returns the array of examples that the plugin can choose for first-time users.
		 */
		private function get_first_time_example_filter() {
			return array(
				'any' => array(
					array(
						'type'      => 'post',
						'default'   => array(
							'post_title'    => array(
								'type'  => 'string',
								'alias' => 'post_title',
							),
							'comment_count' => array(
								'type'  => 'number',
								'alias' => 'comment_count',
							),
						),
					),
					array(
						'type'      => 'post',
						'default'   => array(
							'day(post_date)'    => array(
								'type'  => 'number',
								'alias' => 'day(post_date)',
							),
							'sum(comment_count)'    => array(
								'type'  => 'number',
								'alias' => 'sum(comment_count)',
							),
						),
					),
					array(
						'type'      => 'post',
						'default'   => array(
							'week(post_date)'   => array(
								'type'  => 'number',
								'alias' => 'week(post_date)',
							),
							'sum(comment_count)'    => array(
								'type'  => 'number',
								'alias' => 'sum(comment_count)',
							),
						),
					),
					array(
						'type'      => 'post',
						'default'   => array(
							'month(post_date)'  => array(
								'type'  => 'number',
								'alias' => 'month(post_date)',
							),
							'sum(comment_count)'    => array(
								'type'  => 'number',
								'alias' => 'sum(comment_count)',
							),
						),
					),
				),
			);
		}

		/**
		 * Returns the array of examples that the plugin can choose for first-time users.
		 *
		 * @param   string $feature    The feature for which example is required.
		 * @param   string $type       The chart type for which example is required.
		 */
		private function get_first_time_example( $feature, $type ) {
			$examples   = null;
			switch ( $feature ) {
				case 'filter':
					$examples   = $this->get_first_time_example_filter();
					break;
			}

			if ( ! $examples ) {
				return $examples;
			}

			// choose one random example from the examples.
			if ( array_key_exists( $type, $examples ) ) {
				return $examples[ $type ][ rand( 0, count( $examples ) - 1 ) ];
			}
			return $examples['any'][ rand( 0, count( $examples ) - 1 ) ];
		}

		/**
		 * Checks if the given feature is being used for the first time.
		 *
		 * @param   string $feature    The feature for which we are testing usage.
		 */
		private function is_first_time( $feature ) {
			switch ( $feature ) {
				case 'filter':
					$query  = new WP_Query(
						array(
							'post_type'                 => 'visualizer',
							'no_rows_found'             => false,
							'post_per_page'             => 1,
							'fields'                    => 'ids',
							'update_post_meta_cache'    => false,
							'update_post_term_cache'    => false,
							'meta_query'                => array(
								array(
									'key'       => 'visualizer-filter-config',
									'compare'   => 'EXISTS',
								),
							),
						)
					);
					return ! $query->have_posts();
					break;
			}
			return true;
		}

		/**
		 * Adds the filter creator form
		 */
		public function _addFilterWizard( $chart_id ) {
			if ( ! $this->hasFreeInstalled ) {
				return;
			}
			$hints          = array();
			$filter_config = get_post_meta( $chart_id, 'visualizer-filter-config', true );
			if ( $filter_config ) {
				$hints[]    = __( 'If you get an error while displaying the chart, try to reverse the order of the columns.', 'visualizer' );
			}

			if ( ! $filter_config && $this->is_first_time( 'filter' ) ) {
				$chart_type     = get_post_meta( $chart_id, 'visualizer-chart-type', true );
				// let's check if this is the first time this user is using this feature. if yes, then we will populate the filters with some example data.
				$filter_config  = $this->get_first_time_example( 'filter', $chart_type );
				if ( $filter_config ) {
					$hints[]        = __( 'Since this is the first time you are using this feature, we have added an example filter for you to use. Click on "Show Chart" to see the results', 'visualizer' );
				}
			}

			$default_keys  = null;
			$custom_keys   = null;
			$post_type     = null;
			if ( $filter_config && isset( $filter_config['default'] ) ) {
				$default_keys = $filter_config['default'];
			}
			if ( $filter_config && isset( $filter_config['custom'] ) ) {
				$custom_keys = $filter_config['custom'];
			}
			if ( $filter_config && isset( $filter_config['type'] ) ) {
				$post_type = $filter_config['type'];
			}
			?>
			<div id="filter-wizard" class="visualizer-editor-lhs" style="display: none">
				<div class="filter-wizard-container">
					<div id="filter-wizard-rules">
						<select id="visualizer-post-types">
							<option value=""><?php _e( 'Select post type', 'visualizer' ); ?></option>
							<?php
							$post_types = get_post_types(
								array(
									'public'   => true,
								), 'names'
							);
							foreach ( $post_types as $type ) {
								?>
								<option value="<?php echo $type; ?>" <?php echo $post_type === $type ? 'selected' : ''; ?>><?php echo $type; ?></option>
								<?php
							}
							?>
						</select>
						<select id="visualizer-post-fields" class="visualizer-chosen" multiple
								data-placeholder="<?php _e( 'Select properties', 'visualizer' ); ?>">
							<?php
							$fields = null;
							if ( $post_type ) {
								$fields = $this->get_post_type_properties( $post_type );
								if ( $fields ) {
									if ( is_null( $custom_keys ) ) {
										$custom_keys    = array();
									}
									$selected = array_merge( array_keys( $default_keys ), array_keys( $custom_keys ) );
									foreach ( $fields as $field ) {
										?>
										<option value="<?php echo $field; ?>" <?php echo in_array( $field, $selected ) ? 'selected' : ''; ?>><?php echo $field; ?></option>
										<?php
									}
								}
							}
							?>
						</select>
					</div>
					<div id="vz-no-fields"><?php _e( 'Please select your post type fields which you would like to plot.', 'visualizer' ); ?> </div>
					<div id="filter-wizard-details">
						<form id="filter-wizard-form">
							<input type="hidden" name="post_type" id="filter-wizard-post-type" value="<?php echo $post_type; ?>">
							<table class="filter-wizard">
								<tr class="filter-wizard-template " style="display: none">
									<td></td>
									<td><input type="hidden" name="key[]"><input type="text" name="alias[]"></td>
									<td>
										<select name="type[]">
											<?php foreach ( $this->getAllowedTypes() as $type ) { ?>
												<option value="<?php echo $type; ?>"><?php echo $type; ?></option>
											<?php } ?>
										</select>
									</td>
								</tr>
								<tr>
									<th><?php _e( 'Field', 'visualizer' ); ?></th>
									<th><?php _e( 'Alias', 'visualizer' ); ?></th>
									<th><?php _e( 'Type', 'visualizer' ); ?></th>
								</tr>
								<?php
								if ( ! empty( $default_keys ) ) {
									foreach ( $default_keys as $key => $array ) {
										?>
										<tr class="filter-wizard-new vz-filter-row" id="<?php echo $key; ?>">
											<td><?php echo $key; ?></td>
											<td><input type="hidden" name="key[]" value="<?php echo $key; ?>"><input
														type="text" name="alias[]" value="<?php echo $array['alias']; ?>">
											</td>
											<td>
												<select name="type[]">
													<?php foreach ( $this->getAllowedTypes() as $type ) { ?>
														<option value="<?php echo $type; ?>" <?php echo $type === $array['type'] ? 'selected' : ''; ?>><?php echo $type; ?></option>
													<?php } ?>
												</select>
											</td>
										</tr>
										<?php
									}
								}
								if ( ! empty( $custom_keys ) ) {
									foreach ( $custom_keys as $key => $array ) {
										?>
										<tr class="filter-wizard-new vz-filter-row" id="<?php echo $key; ?>">
											<td><?php echo $key; ?></td>
											<td><input type="hidden" name="key[]" value="<?php echo $key; ?>"><input
														type="text" name="alias[]" value="<?php echo $array['alias']; ?>">
											</td>
											<td>
												<select name="type[]">
													<?php foreach ( $this->getAllowedTypes() as $type ) { ?>
														<option value="<?php echo $type; ?>" <?php echo $type === $array['type'] ? 'selected' : ''; ?>><?php echo $type; ?></option>
													<?php } ?>
												</select>
											</td>
										</tr>
										<?php
									}
								}
								?>
							</table>
						</form>
					</div>
					<ul class="filter-wizard-hints">
					<?php
					if ( $hints ) {
						foreach ( $hints as $hint ) {
					?>
					<li><?php echo $hint; ?></li>
					<?php
						}
					}
					?>
					</ul>
				</div>
			</div>
			<?php
		}

		/**
		 * Update hook
		 *
		 * @param mixed $series Series options.
		 * @param mixed $data Chart data options.
		 */
		public function _addUpdateHook( $series, $data ) {
			echo 'if (win.visualizer1) {';
			echo 'win.visualizer1.render(', $series, ',', $data, ');';
			echo '}';
		}

		/**
		 * Gets the properties of the post type
		 *
		 * @access friendly
		 */
		function fetch_post_type_properties() {
			$nonce = isset( $_POST['nonce'] ) ? wp_verify_nonce( $_POST['nonce'], self::ACTION_FETCH_POST_TYPE_PROPS ) : '';
			if ( $nonce ) {
				$array = $this->get_post_type_properties( $_POST['post_type'] );
				$this->_sendResponse(
					array(
						'success' => true,
						'fields'  => $array,
					)
				);
				wp_die();
			}
		}

		/**
		 * Gets the properties of the post type
		 *
		 * @access private
		 */
		private function get_post_type_properties( $post_type ) {
			$array = null;
			$query = new WP_Query(
				array(
					'post_type'              => $post_type,
					'no_rows_found'          => false,
					'post_per_page'          => 1,
					'orderby'                => 'post_date',
					'order'                  => 'DESC',
					'fields'                 => 'ids',
					'update_post_meta_cache' => false,
					'update_post_term_cache' => false,
				)
			);
			$array = array();
			foreach ( self::$POST_DEFAULT_PROPERTIES as $name => $type ) {
				$array[] = $name;
				switch ( $type ) {
					case 'int':
						$array[]    = "sum($name)";
						$array[]    = "avg($name)";
						break;
					case 'string':
						$array[]    = "count($name)";
						break;
					case 'date':
						$array[]    = "day($name)";
						$array[]    = "week($name)";
						$array[]    = "month($name)";
						$array[]    = "year($name)";
						break;
				}
			}

			if ( $query->have_posts() ) {
				$id   = $query->posts[0];
				$meta = get_post_meta( $id, '', true );
				foreach ( $meta as $key => $values ) {
					$array[] = $key;
					$array[] = "count($key)";
				}
			}

			return $array;
		}

		/**
		 * Determines the db field and the function used from the input parameter.
		 *
		 * @param string $key    The input parameter.
		 * @return array
		 */
		private function get_column_and_function( $key ) {
			preg_match( '/(.*)\((.*)\)/', $key, $output );
			if ( count( $output ) === 3 ) {
				return array( $output[2], $output[1] );
			}
			return $key;
		}

		/**
		 * Gets the data of the selected post type/properties
		 *
		 * @access friendly
		 */
		function fetch_post_type_data() {
			global $wpdb;
			$post_columns = array_keys( self::$POST_DEFAULT_PROPERTIES );
			$nonce    = isset( $_POST['nonce'] ) ? wp_verify_nonce( $_POST['nonce'], self::ACTION_FETCH_POST_TYPE_DATA ) : '';
			$chart_id = isset( $_POST['chart_id'] ) ? filter_var( $_POST['chart_id'], FILTER_VALIDATE_INT ) : '';
			if ( $nonce && $chart_id ) {
				$custom_keys  = array();
				$default_keys = array();
				parse_str( $_POST['props'], $params );
				$params = array_filter( $params );
				$index  = 0;
				$aggregate_function = false;
				// this will capture the order the elements are sent as this is very important. The first element is the x axis and second is the y axis
				$order = array();
				foreach ( $params['key'] as $field ) {
					if ( ! $field ) {
						$index ++;
						continue;
					}
					$key            = $field;
					$function       = '';
					$col_function   = $this->get_column_and_function( $field );

					if ( is_array( $col_function ) ) {
						$key        = $col_function[0];
						$function   = $col_function[1];
						if ( ! $aggregate_function && in_array( $function, self::$AGGREGATOR_FUNCTIONS ) ) {
							$aggregate_function = true;
						}
					}
					$order[] = $key;
					if ( in_array( $key, $post_columns ) ) {
						$default_keys[ $field ] = array(
							'type'  => $params['type'][ $index ],
							'alias' => $params['alias'][ $index ],
						);
					} else {
						$custom_keys[ $key ] = array(
							'type'  => $params['type'][ $index ],
							'alias' => $params['alias'][ $index ],
						);
					}
					$index ++;
				}
				$query  = null;
				$values = array();
				// this will capture the first element in the select order
				$first = array_shift( $order );
				$group_by = '';
				if ( $aggregate_function ) {
					$group_by = ' GROUP BY id ';
				}

				if ( count( $custom_keys ) > 1 ) {
					// both custom keys
					$results = $wpdb->get_col( $wpdb->prepare( "SELECT id FROM {$wpdb->prefix}posts WHERE post_type=%s LIMIT 1000", $params['post_type'] ) );
					if ( $results ) {
						$keys   = implode( ',', array_fill( 0, count( $custom_keys ), '%s' ) );
						$ids    = implode( ',', array_fill( 0, count( $results ), '%d' ) );
						$query  = "SELECT meta_value FROM {$wpdb->prefix}postmeta WHERE meta_key IN ($keys) AND post_id IN ($ids) $group_by ORDER BY post_id, FIELD(meta_key, '" . implode( '\',\'', array_keys( $custom_keys ) ) . "')";
						$values = array_merge( array_keys( $custom_keys ), $results );
					}
				} elseif ( 1 === count( $custom_keys ) ) {
					// one custom key and one post property
					// lets check if the first field is a custom key or a property so that the select statement can be created in the same order
					$select = '';
					if ( in_array( $first, $post_columns ) ) {
						$select = implode( ',', array_keys( $default_keys ) ) . ', pm.meta_value';
					} else {
						$select = 'pm.meta_value, ' . implode( ',', array_keys( $default_keys ) );
					}
					$query    = "SELECT $select FROM {$wpdb->prefix}posts p LEFT JOIN {$wpdb->prefix}postmeta pm ON p.id = pm.post_id and pm.meta_key=%s WHERE p.post_type=%s $group_by LIMIT 1000";
					$values[] = implode( ',', array_keys( $custom_keys ) );
					$values[] = $params['post_type'];
				} else {
					// both post properties
					$query    = 'SELECT ' . implode( ',', array_keys( $default_keys ) ) . " FROM {$wpdb->prefix}posts WHERE post_type=%s $group_by LIMIT 1000";
					$values[] = $params['post_type'];
				}

				$data = array();
				if ( $query ) {
					$query  = $wpdb->prepare( $query, $values ); // WPCS: unprepared SQL ok.
					$results = $wpdb->get_results( $query, ARRAY_N ); // WPCS: unprepared SQL ok.
					if ( $results ) {
						$aliases = array();
						$types   = array();
						// lets see what the order of default/custom keys are for processing
						$_keys = array_merge( $default_keys, $custom_keys );
						if ( ! in_array( $first, $post_columns ) ) {
							$_keys = array_merge( $custom_keys, $default_keys );
						}
						foreach ( $_keys as $key => $array ) {
							$aliases[] = $array['alias'];
							$types[]   = $array['type'];
						}
						$data[] = $aliases;
						$data[] = $types;
						if ( count( $custom_keys ) > 1 ) {
							$row   = array();
							$index = 0;
							foreach ( $results as $field ) {
								$row[] = $field[0];
								// because the meta_values will be in one column and will alternate, every odd result will complete the "record"
								// this assumes that each post will have both the meta_keys
								if ( 0 !== $index ++ % 2 ) {
									$data[] = $row;
									$row    = array();
								}
							}
						} else {
							foreach ( $results as $row ) {
								$data[] = $row;
							}
						}
					}
				}
				if ( $data ) {
					/*
                    // triggering uploadCharts by manually triggering action does not work because the script dies and thus the front end rendering does not happen
                    // even if we add an action that internally calls uploadCharts, it does not work
                    $_GET                   = array();
                    $_POST                  = array();
                    $_GET['nonce']          = wp_create_nonce();
                    $_GET['chart']          = $chart_id;
                    $_POST['chart_data']    = json_encode( $data );
                    if ( ! defined( 'DOING_AJAX' ) ) {
                        define( 'DOING_AJAX', true );
                    }
                    do_action( 'wp_ajax_' . Visualizer_Plugin::ACTION_UPLOAD_DATA );
                    */
					update_post_meta(
						$chart_id, 'visualizer-filter-config', array(
							'default' => $default_keys,
							'custom'  => $custom_keys,
							'type'    => $params['post_type'],
						)
					);
				}
				$this->_sendResponse(
					array(
						'success' => true,
						'data'    => json_encode( $data ),
					)
				);
				wp_die();
			}// End if().
		}

		/**
		 * Gets the chart data corresponding to a specific chart
		 *
		 * @since 1.4.2.3
		 *
		 * @access public
		 */
		public function fetchData() {
			$nonce    = isset( $_GET['nonce'] ) ? wp_verify_nonce( $_GET['nonce'], Visualizer_Pro::ACTION_FETCH_DATA ) : '';
			$chart_id = isset( $_GET['chart_id'] ) ? filter_var( $_GET['chart_id'], FILTER_VALIDATE_INT ) : '';
			if ( $chart_id ) {
				$chart   = get_post( $chart_id );
				$success = $chart && $chart->post_type == Visualizer_Plugin::CPT_VISUALIZER;
				$type    = get_post_meta( $chart->ID, Visualizer_Plugin::CF_CHART_TYPE, true );
				$series  = apply_filters( Visualizer_Plugin::FILTER_GET_CHART_SERIES, get_post_meta( $chart->ID, Visualizer_Plugin::CF_SERIES, true ), $chart->ID, $type );
				$data    = apply_filters( Visualizer_Plugin::FILTER_GET_CHART_DATA, unserialize( $chart->post_content ), $chart->ID, $type );
				if ( $success ) {
					$results = array(
						'success' => true,
						'series'  => $series,
						'data'    => $data,
					);
					$this->_sendResponse( $results );
				}
			}
			wp_die();
		}

		/**
		 * Sends the response to the ajax requester
		 *
		 * @access private
		 */
		private function _sendResponse( $results ) {
			$methodChecker = new ReflectionMethod( 'Visualizer_Module_Chart', '_sendResponse' );
			if ( $methodChecker->isStatic() ) {
				Visualizer_Module_Chart::_sendResponse( $results );
			} else {
				header( 'Content-type: application/json' );
				nocache_headers();
				echo json_encode( $results );
			}
		}

		/**
		 * Gets the allowed types
		 *
		 * @access private
		 */
		private function getAllowedTypes() {
			return method_exists( 'Visualizer_Source', 'getAllowedTypes' ) ? Visualizer_Source::getAllowedTypes() : array(
				'string',
				'number',
				'boolean',
				'date',
				'datetime',
				'timeofday',
			);
		}

		/**
		 * Write debug messages
		 *
		 * @param string $msg Message to report.
		 */
		static function writeDebug( $msg ) {
			if ( Visualizer_Pro_Debug ) {
				try {
					mkdir( Visualizer_Pro_PATH . '/tmp' );
				} catch ( Exception $e ) {
					error_log( 'Unable to create report dir' );
				}
				// @unlink(Visualizer_Pro_PATH . "/tmp/log.log");
				file_put_contents( Visualizer_Pro_PATH . '/tmp/log.log', date( 'F j, Y H:i:s' ) . ' - ' . $msg . "\n", FILE_APPEND );
			}
		}

		/**
		 * Return if the user is a business user
		 *
		 * @param string $dumb Test data.
		 *
		 * @return boolean
		 */
		function visualizer_is_business( $dumb = null ) {
			$plan = get_option( 'visualizer_pro_license_plan', 1 );
			$plan = intval( $plan );
			return ( $plan > 1 );
		}

		/**
		 * Adds ajax action for the pro
		 *
		 * @param mixed $plugin Free plugin class.
		 *
		 * @return array
		 */
		public function _getAjaxAction( $plugin ) {
			return array( Visualizer_Pro::ACTION_FETCH_DATA, 'fetchData', $this );
		}

		/**
		 * Add form elements
		 */
		public function _addFormElements() {
			if ( ! $this->hasFreeInstalled ) {
				return;
			}
			echo '<input type="hidden" id="chart-data" name="chart_data">';
		}

		/**
		 * Add editor elements
		 */
		public function _addEditorElements() {
			if ( ! $this->hasFreeInstalled ) {
				return;
			}
			$fetch_link = add_query_arg(
				array(
					'action' => Visualizer_Pro::ACTION_FETCH_DATA,
					'nonce'  => wp_create_nonce(),
				), admin_url( 'admin-ajax.php' )
			);
			?>
			<div id="editor-resources" data-ajax="<?php echo $fetch_link; ?>"></div>
			<?php
			$query_args_charts = array(
				'post_type'      => Visualizer_Plugin::CPT_VISUALIZER,
				'posts_per_page' => - 1,
			);
			$charts            = array();
			$query             = new WP_Query( $query_args_charts );
			?>
			<input type="button" class="button from-chart-btn" id="existing-chart" name="existing_chart" value="<?php esc_attr_e( 'From Chart', 'visualizer' ); ?>:">
			<select id="chart-id" name="chart_id">
				<?php
				while ( $query->have_posts() ) {
					$chart    = $query->next_post();
					$settings = get_post_meta( $chart->ID, Visualizer_Plugin::CF_SETTINGS, true );
					?>
					<option value="<?php echo $chart->ID; ?>"><?php echo empty( $settings['title'] ) ? '#' . $chart->ID : $settings['title']; ?></option>
					<?php
				}
				?>
			</select>
			<div>
				<button id="editor-chart-button" class="button button-primary show-live-btn" data-current="chart"
						data-t-chart="<?php esc_attr_e( 'Show Live Editor', 'visualizer' ); ?>"
						data-t-editor="<?php esc_attr_e( 'Show Chart', 'visualizer' ); ?>"><?php esc_attr_e( 'Show Live Editor', 'visualizer' ); ?></button>
			</div>

			<?php
		}
	}
} // End if().
if ( class_exists( 'Visualizer_Pro' ) ) {
	global $Visualizer_Pro;
	$Visualizer_Pro = new Visualizer_Pro();
}
